from django.apps import AppConfig
from trim.apps import live_import
from . import context

from django.conf import settings

class ThemingConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'trim.theming'

    def ready(self):
        context.build_default_themer()
        # Import /app/theming.py from all apps
        # print(' -- live importing theming')
        live_import('theming')



from django import template
from django.urls import resolve, reverse


register = template.Library()

quickform_kwargs = ('form_submit_label', 'form_submit_button', )

@register.inclusion_tag('trim/quickform.html', takes_context=True, name='quickform.form')
def quickform_template(context, view_name, *args, **kwargs):
    """Apply a quickform within a form template, utilising the quickform as
    the thing to render.

        {% quickform.form 'associations:grouping-search-form'
            form_submit_button=True
            form_submit_label='Search'
            id='quicksearchform'
        %}

    any option not in `quickform_kwargs` is applied to the form a node parameter.
    """
    form = quickform(context, view_name, *args, **kwargs)
    extra = { 'opts': {}, 'trim_form': {}}

    for x,v in kwargs.items():
        # x, y for x, y in kwargs.items() if x in quickform_kwargs
        k = ['opts', 'trim_form'][bool(x in quickform_kwargs)]
        extra[k][x] = v

    return {
        'trim_quickform_form': form,
        **extra,
    }


@register.simple_tag(takes_context=True, name='quickform')
def quickform(context, view_name, *url_args, **kwargs):
    """Build a form given the view_name and any `view_arguments`

        {% quickform 'associations:grouping-search-form' as grouping_search_form %}
        <form id='grouping_search_form'
            method='post'
            action='{{ grouping_search_form.action_url }}'>
            {{ grouping_search_form.as_ul }}
            {% csrf_token %}
            <button type='submit'>Submit</button>
        </form>

    """
    request = context.get('request')
    args = kwargs.get('view_args', ()) if len(url_args) == 0 else url_args


    rev = reverse(view_name, args=args)
    resolve_match = resolve(rev)

    view = resolve_match.func.view_class
    initkwargs = resolve_match.func.view_initkwargs
    view.view_initkwargs = initkwargs

    instance = view(**initkwargs)
    instance.setup(request, *args, **kwargs)

    form = instance.get_form()
    form.action_url = rev
    initial = kwargs.get('initial', None) or kwargs
    if hasattr(form, 'initial') and (len(initial) > 0):
        form.initial.update(initial)
    return form




VIEWS = (

    ## crud
    'ListView',
    'CreateView',
    'UpdateView',
    'DeleteView',
    'DetailView',
    # 'DeleteViewCustomDeleteWarning',

    ## history
    'ArchiveIndexView',
    'DateDetailView',
    'DayArchiveView',
    'MonthArchiveView',
    'TodayArchiveView',
    'WeekArchiveView',
    'YearArchiveView',

    'RedirectView',
    'TemplateView',

    'FormView',
    'LoginView',
    'LogoutView',

    'PasswordChangeDoneView',
    'PasswordChangeView',
    'PasswordResetCompleteView',
    'PasswordResetConfirmView',
    'PasswordResetDoneView',
    'PasswordResetView',

)

## Order is importtant.
url_enforcements = (
    ( ('update', 'detail', 'delete',), '{name}/<str:pk>/', ),
    # ( ('archiveindex',), '', ),
    ( ('todayarchive',), 'today/', ),
    ( ('datedetail',  ), '<int:year>/<str:month>/<int:day>/<int:pk>/', ),
    ( ('dayarchive',  ), '<int:year>/<str:month>/<int:day>/', ),
    ( ('montharchive',), '<int:year>/<str:month>/' , ),
    ( ('weekarchive', ), '<int:year>/week/<int:week>/', ),
    ( ('yeararchive', ), '<int:year>/', ),
)

## Generate a Mapped name lise of the VIEWS, producing a list for direct
## name association:
    # 'list':'ListView',
    # 'create':'CreateView',
    # 'update':'UpdateView',
    # 'delete':'DeleteView',
    # 'detail':'DetailView',
MAPPED_NAMES = { x[:-4].lower(): x for x in VIEWS }

## Write the reverse of the MAPPED_NAMES, for class to friendly map:
    # ListView: list
MAPPED_CLASS = {y:x for x,y in MAPPED_NAMES.items()}


def crud():
    return (
        'list',
        'create',
        'update',
        'delete',
        'detail',
        )

def history():
    # Order is important
    return (
        'archiveindex',
        'todayarchive',
        'weekarchive',
        'datedetail',
        'dayarchive',
        'montharchive',
        'yeararchive',
    )

def tidy_enforcements(enforcements):
    url_parts = {}
    for names, enforced in enforcements:
        for n in names:
            url_parts[n] = enforced
    return url_parts


def render_defaults(names, parts):
    refs = {}
    res = {}
    for name, classname in names.items():
        refs['name'] = name
        refs['classname'] = classname
        live_part = parts.get(name, None)
        first = f'{name}/'
        part = ''
        if live_part is not None:
            part = live_part.format(**refs)
            first = ''
        res[name] = f'{first}{part}'
    return res


URL_PARTS = tidy_enforcements(url_enforcements)
URL_DEFAULTS = render_defaults(MAPPED_NAMES, URL_PARTS)

def get_url(name):
    return URL_DEFAULTS.get(name)

import inspect

def get_mapped_name(instance):
    mro = inspect.getmro(instance.__class__)
    for _c in mro:
        r = MAPPED_CLASS.get(_c.__name__)
        if r is not None:
            return r



from django.contrib.auth.mixins import UserPassesTestMixin

def is_staff_or_admin(user):
     return user.is_superuser or user.is_staff


class IsStaffMixin(UserPassesTestMixin):
    """
    Allow if the user is staff or admin.
    """
    def test_func(self):
        uu = self.request.user
        active_staff = uu.is_active and is_staff_or_admin(uu)
        return active_staff


class MissingField(Exception):
    pass


class UserOwnedMixin(UserPassesTestMixin):
    """
    Ensure the user owning the inner model `get_object()` is the requesting user.

        from trim.views import UserOwnedMixin
        class AddressDetailView(UserOwnedMixin, shorts.DetailView):
            model = models.Address
            user_field = 'creator'
            user_allow_staff = True
            template_name = 'locality/address_detail.html'
    """
    user_field = 'user'
    user_allow_staff = False

    def test_func(self):
        uu = self.request.user
        request_pk = uu.pk
        model_user = getattr(self.get_object(), self.user_field)

        try:
            owns = model_user.pk == request_pk
        except AttributeError as err:
            raise MissingField(self.user_field) from err

        is_admin = is_staff_or_admin(uu) if self.user_allow_staff else False
        active_staff = uu.is_active and (owns or is_admin)

        return active_staff



from wagtail.images import get_image_model_string
from wagtail.core.fields import StreamField

from wagtail.core.models import Orderable
from modelcluster.fields import ParentalKey

from wagtail.core import blocks
from wagtail.images.blocks import ImageChooserBlock

from .django import fk
from .base import defaults

from .. import panels
from wagtail.snippets.models import register_snippet

register = register_snippet

def streamfield(*a, **kw):
    kw = defaults(a, kw, nil=True)
    return StreamField(*a, **kw)


def image_fk(*a, **kw):
    kw = defaults(a, kw, related_name='+')
    return fk(get_image_model_string(), *a, **kw)

image = image_fk



from django.template import Context
from django import template
from django.template.base import token_kwargs

from django.forms.boundfield import BoundField
from trim import get_model

register = template.Library()


@register.simple_tag(takes_context=True)
def stockcount_product(context, stockcount_id):
    SC = get_model('stocks.StockCount')
    if isinstance(stockcount_id, BoundField):
        stockcount_id = stockcount_id.value()

    if stockcount_id == None: return ''

    return SC.objects.get(id=stockcount_id).product

# @register.inclusion_tag('dummy.html')
# def wrap(template='default.html'):
#     return {'template': template}



"""

Experiments:
    res = self.run_poll_command(command)

    res = self.run_command(command)

    subcall_stream(command)

    read_one_stream_command(command)

    gen = self.run_command2(command)
    for line in iter(gen):
        print(line, end='')

Raw method:
    print(res)
    res = subprocess.run(command, shell=True, capture_output=True)
    print('  code: ', res.returncode)
    out = clean(res.stdout)
    err = clean(res.stderr)
    print('  out:  ', out)
    print('  error:', err)
    print(' ', dir(res))

"""
import subprocess
import sys

import pkg_resources


def main():
    print('Inject')

__requires__ = 'trim'
import sys
from pkg_resources import load_entry_point

if __name__ == '__main__':
    sys.exit(
        load_entry_point('django-trim==0.1', 'console_scripts', 'trim')()
    )

def inj():
    # Create the fake entry point definition
    ep = pkg_resources.EntryPoint.parse('dummy = dummy_module:DummyPlugin')

    # Create a fake distribution to insert into the global working_set
    d = pkg_resources.Distribution()

    # Add the mapping to the fake EntryPoint
    d._ep_map = {'console_scripts': {'dummy': ep}}

    # Add the fake distribution to the global working_set
    pkg_resources.working_set.add(d, 'dummy')


def test_entry_point():
    distribution = pkg_resources.Distribution(__file__)
    entry_point = pkg_resources.EntryPoint.parse(
        'plugin1=plugins.plugin1:plugin1_class',
        dist=distribution)
    distribution._ep_map = {'my_project.plugins': {'plugin1': entry_point}}
    pkg_resources.working_set.add(distribution)


def subcall_stream(cmd, fail_on_error=True):
    # Run a shell command, streaming output to STDOUT in real time
    # Expects a list style command, e.g. `["docker", "pull", "ubuntu"]`
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        universal_newlines=True)
    for line in p.stdout:
        sys.stdout.write(line)
    p.wait()


    return exit_code

def read_one_stream_command(command, show=True):
    process = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        ## not required when processing by read(1)
        # bufsize=1,
        ## Adding universal newlines will evoke _str_ from the `out`
        # universal_newlines=True,
    )
    res = ''
    byte_count = 0
    count = 0
    while True:
        out = process.stdout.read(1)
        if out == '' and process.poll() != None:
            break
        byte_count += 1

        if byte_count > 75_000:
            break

        if len(out) > 0:
            t = out.decode('utf')
            res += t
            if show:
                sys.stdout.write(t)
                sys.stdout.flush()
        else:
            count += 1
            if count % 20 == 0:
                print('blanked.')
                break
    # process.wait()
    return res


def run_poll_command(command):
    process = subprocess.Popen(command, stdout=subprocess.PIPE)
    while True:
        output = process.stdout.readline()
        if output == '' and process.poll() is not None:
            break
        if output:
            print(output.strip())
    rc = process.poll()
    return rc


def run_command(command):
    try:

        process = subprocess.Popen(command,
                        stdout=subprocess.PIPE,
                        stdin=subprocess.PIPE,
                        stderr=subprocess.STDOUT,
                        shell=True,
                        encoding='utf8')
        while True:
            inner = True
            pack = ''
            while inner:
                char = process.stdout.readline(1)
                pack += char
                if char == '\n':
                    print(pack, end='')
                if char is None:
                    print(pack)
                inner = False
            print(pack, end='')
            output = process.stdout.readline(64)
            if output == '' and process.poll() is not None:
                print("no output")
                break
            if output:
                print(output.strip())
        rc = process.poll()
        return rc
    except KeyboardInterrupt:
        # process.terminate()
        print('KeyboardInterrupt')


def run_command2(cmd):
    popen = subprocess.Popen(cmd,
            stdout=subprocess.PIPE,
            # stdin=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
            shell=True,
            encoding='utf8',
            bufsize=1,
            )
    stdout = popen.stdout
    for stdout_line in iter(stdout.readline, ""):
        yield stdout_line
        stdout.flush()

    stdout.flush()
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd)
    return return_code

if __name__ == '__main__':
    main()





DEFAULT_NIL = True


def defaults(args, params, nil_sub=True, nil_key='nil', **kw):

    if nil_sub:
        # nil subtract, or substitution
        #  Nil. being blank+null - where nil=True
        if ('nil' in kw) or ('nil' in params):
            val = kw.get('nil', None) or params.get('nil', None)

            if isinstance(val, bool):
                # if the given value is a single bool,
                # unpack and multiply into the count of
                # args for the blank_null func,
                val = [val] * 2
                kw.update(**blank_null(*val))

    for k,v in kw.items():
        if k == nil_key:
            continue
        params.setdefault(k, v)
    params.pop('nil', None)
    return params


def blank_null(b=True, n=True):
    return {'blank':b, 'null': n}



"""
   {% timedelta cart.updated cart.created format="" %}
   {{ cart.created|timesince:cart.updated }}
   {% human_timedelta cart.updated cart.created format="" %}
"""
from django import template
from datetime import timedelta
from django.urls import reverse
from django.utils.translation import ngettext

register = template.Library()



@register.simple_tag(takes_context=False, name='timedelta')
def timedelta_tag(late_time, early_time, *targs, **kwargs):
    td = (late_time - early_time)
    return td

@register.simple_tag(takes_context=False, name='human_timedelta')
def str_timedelta_tag(late_time, early_time, *targs, **kwargs):
    td = (late_time - early_time)
    return localize_timedelta(td)


def localize_timedelta(delta):
    ret = []
    num_years = int(delta.days / 365)
    if num_years > 0:
        delta -= timedelta(days=num_years * 365)
        ret.append(ngettext('%d year', '%d years', num_years) % num_years)

    if delta.days > 0:
        ret.append(ngettext('%d day', '%d days', delta.days) % delta.days)

    num_hours = int(delta.seconds / 3600)
    if num_hours > 0:
        delta -= timedelta(hours=num_hours)
        ret.append(ngettext('%d hour', '%d hours', num_hours) % num_hours)

    seconds= delta.seconds
    num_minutes = int(seconds / 60)
    if num_minutes > 0:
        ret.append(ngettext('%d minute', '%d minutes', num_minutes) % num_minutes)
    else:
        ret.append(ngettext('%d second', '%d seconds', seconds) % seconds)
    return ' '.join(ret)



from django.urls import reverse
from django import template

register = template.Library()

from ..urls import absolute_reverse, absolutify



@register.simple_tag(takes_context=False, name='link.info')
def link_info(view_name, *targs, **kwargs):
    return link(view_name, *targs, **kwargs)


@register.inclusion_tag('trim/link.html', takes_context=True)
def abs_link(context, link, *targs, **kwargs):
    res = gen_link(link, *targs, **kwargs)
    res['url'] = absolutify(context['request'], res['url'])
    return res


@register.inclusion_tag('trim/link.html', takes_context=False)
def link(link, *targs, **kwargs):
    if isinstance(link, ShadowDict):
        if len(targs) > 0:
            link['text'] = targs[0]
        link['kwargs'].update(kwargs)
        return link
    return gen_link(link, *targs, **kwargs)


@register.inclusion_tag('trim/link.html', takes_context=False)
def new_link(link, *targs, **kwargs):
    kwargs.setdefault('target', '_blank')
    kwargs.setdefault('rel', 'noreferrer noopener')

    return gen_link(link, *targs, **kwargs)


@register.inclusion_tag('trim/js_link.html', name='js', takes_context=False)
def script_link(link, *targs, **kwargs):
    kwargs.setdefault('type', 'text/javascript')
    return {
        'static_name': link,
        'kwargs': kwargs,
    }


@register.inclusion_tag('trim/css_link.html', name='css', takes_context=False)
def css_link(link, *targs, **kwargs):
    """

    Build a standard <link> for CSS, using a static path

        {% css 'css/main.css' %}

    All attributes of the node are available. As an example of versatility
    (not recommended):

        {% css 'images/favicon.ico' rel="shortcut icon" type="image/png" %}
    """
    kwargs.setdefault('rel', 'stylesheet')
    kwargs.setdefault('type', 'text/css')
    return {
        'static_name': link,
        'kwargs': kwargs,
    }


@register.inclusion_tag('trim/link.html', takes_context=False)
def new_url_link(link, *targs, **kwargs):
    """Provide a _real_ URL, rather than a reversable name:

        {% load link %}
        {% for pay_url in flow.cart.receipt_links %}
            {% new_url_link pay_url "Receipt" %}
        {% endfor %}

    """
    kwargs.setdefault('target', '_blank')
    kwargs.setdefault('rel', 'noreferrer noopener')
    kwargs.setdefault('url', link)
    return gen_link(link, *targs, **kwargs)


@register.inclusion_tag('trim/link.html', takes_context=False)
def url_link(link, *targs, **kwargs):
    """Provide a _real_ URL, rather than a reversable name:

        {% load link %}
        {% for pay_url in flow.cart.receipt_links %}
            {% url_link pay_url "Receipt" %}
        {% endfor %}

    """
    # kwargs.setdefault('target', '_blank')
    # kwargs.setdefault('rel', 'noreferrer noopener')
    kwargs.setdefault('url', link)
    return gen_link(link, *targs, **kwargs)


class ShadowDict(dict):
    pass


def gen_link(link, *targs, **kwargs):
    # "...X as name %}" detection
    hotix_as_word = targs[-2:] == ('', '',)
    ## We cut _3_ as (1) for the _last argument (text) and 2 for 'as X'
    ## else -1 for just the _last argument_
    cutlen = -3 if hotix_as_word else -1
    # Args are exerything given except the last "text" arg (and hotfix 'as X')
    args = targs[:cutlen] if len(targs) > 0 else targs

    url = kwargs.get('url', None) or reverse(link, args=args)
    text = targs[cutlen] if len(targs) > 0 else url

    return ShadowDict(
        url=url,
        text=text,
        kwargs=kwargs,
    )




from wagtail.core.blocks import StreamBlock

from wagtailmarkdown.blocks import MarkdownBlock as WM_MarkdownBlock
from trim.wagtail import blocks as t_blocks
from django.forms.widgets import CheckboxSelectMultiple
from django import forms

class Styles(t_blocks.StructBlock):


    styles = t_blocks.MultipleChoiceBlock(
            widget=CheckboxSelectMultiple,
            choices=(
                ('flipped', 'Horizontal Flip',),
                # ('b', 'b',),
            )
        )

    def __init__(self, required=True, help_text=None, **kwargs):
        super().__init__(**kwargs)
        ss=self.declared_blocks['styles']
        ch = kwargs.get('choices', (('egg','Egg'),))
        ss.choices = ch

    class Meta:
        template = 'blocks/styles.html'

from wagtail.core.blocks import StructValue

class LinkStructValue(StructValue):
    """
    https://docs.wagtail.org/en/v4.0.1/advanced_topics/customisation/streamfield_blocks.html#additional-methods-and-properties-on-structblock-values
    """
    def html(self):
        text = self.get('content')
        # page = self.get('page')
        return text or "No text for {self}"


class MarkdownBlock(t_blocks.StructBlock):
    content = StreamBlock((
            ('text', WM_MarkdownBlock(icon="pilcrow"),),
            ('document', t_blocks.document_chooser(icon="doc-full-inverse"),),
        ))

    # text = WM_MarkdownBlock(icon="pilcrow")
    # document = t_blocks.document_chooser(icon="doc-full-inverse")
    css_styles = Styles()

    class Meta:
        template = 'blocks/markdown-block.html'
        value_class = LinkStructValue



from .base import AppActions, Help as BHelp, AppFunction, AppArgument, ConfigMixin, print_help, SubHelpFormatter
from pathlib import Path
from functools import partial
import shlex
from trim import execute

import argparse


# def print_help(parser):

#     # retrieve subparsers from parser
#     subparsers_actions = [
#         action for action in parser._actions
#         if isinstance(action, argparse._SubParsersAction)]
#     # there will probably only be one subparser_action,
#     # but better safe than sorry
#     for subparsers_action in subparsers_actions:
#         # get all subparsers and print help

#         for choice, subparser in subparsers_action.choices.items():
#             print("--- Subparser '{}'".format(choice))
#             subparser.formatter_class = SubHelpFormatter
#             print(subparser.format_help())

class Help(BHelp):
    double_dash = ('apply a double-switch before the '
                    'input to ensure switch aggregation.')
    pattern_append = ('Append a partial string to the filename split pattern')


class GraphApps(object):
    verbose = 0

    def build_graph_parsers(self, top, parser=None, depth=0, position=None):
        """Given a nested tree of parsers build sub apps into the argparser

            {
                "graph": {
                    "keys": [
                        "download",
                    ],
                    "download": {
                        "depth": 0,
                        "keys": ["db"],
                        "position": ["db"],
                        "items": [
                            "download"
                        ],
                        "db": {
                            "depth": 1,
                            "keys": [],
                            "position": [],
                            "items": [
                                "db"
                            ]
                        }
                    }
                },
                "db": "py scripts/download_remote.py"
            }

            trim download db
        """
        for key in top.get('keys', ()):
            data = top[key]
            inner_keys = data.get('keys', ())
            position = tuple(data.get('position', ()))
            position = (position or ()) + (key,)
            if self.verbose > 0:
                w =' ' * depth
                print(w, 'Adding', position)
            # build a caller
            func = partial(self.depthed_default_caller, data)
            # Push into the parsers
            sub_parser = self.add_sub(func, key, parser, formatter_class=SubHelpFormatter)
            # recurse.
            self.build_graph_parsers(data, sub_parser, depth=depth+1, position=position)

    def depthed_default_caller(self, content, args=None):
        items = content.get('items', ())
        conf = self.get_conf()
        datas = {x:conf.get(x) for x in items}
        self.namespace = args
        self.run_units(datas)
        return 'res from default_caller'

    def run_units(self, datas):
        # print('\nKey', datas)
        # print('  default_caller: Key', self.namespace)
        for i, (k, d) in enumerate(datas.items()):
            self.execute_step(i, k, d)

    def execute_step(self, i, k, d):
        pass


class StepExecute(object):

    def execute_step(self, i, k, d):
        # Shell true, because the user is building the scripts.
        command = shlex.split(d)
        command += self.namespace.unknown_args
        print('  Running', k, ' -> ', command)
        execute.read_one_stream_command(command)


class TrimApp(AppActions, StepExecute, GraphApps):
    """TrimApp extends AppActions to present provide trim app functions
    for the CLI

    """
    prog_name = 'trim'
    primary_name = 'trim'

    def setup(self):
        conf = self.get_conf()
        graph = conf.get('graph', {})
        self.verbose = conf.get('verbose', 0)
        self.build_graph_parsers(graph)

    def scripts_func(self, args):
        print('scripts_func', args)
        return 'res from scripts_func'

    def default_caller(self, args):
        print_help(self.get_primary_parser())
        return 'res from default_caller'

    def add(self, *app_function):
        self.app_functions += app_function


class TrimAdminApp(TrimApp):
    register_name = 'admin'

    def get_conf(self):
        return {}


class VerboseSwitch(AppArgument):

    def setup_args(self, parser):
        parser.add_argument('--verbose', '-v', action='count', default=0)


class VersionSwitch(AppArgument):
    _version = '1.0'
    target =  'primary'

    def setup_args(self, parser):
        parser.add_argument('--version',
                    action='version',
                    version=f'%(prog)s {self._version}'
                )


class Scripts(AppFunction):
    # name = 'scripts'
    help = Help.scripts
    register_name ='admin'


class ScriptsAddFilenameArg(AppArgument):
    """The 'filename' argument for the `scripts add` function.

    This is applied through the `arguments` of ScriptsAdd.

    As this defines early, it will fail the primary hook parser.
    """
    target ='scripts.add'

    def setup_args(self, parser):
        parser.add_argument('filename', type=str, nargs='*', help=Help.add_filename, default='.')
        parser.add_argument('--', dest='_', nargs='+', type=str, default=None, help=Help.double_dash)
        parser.add_argument('--type', type=str, default='raw', choices=('raw', 'path',))
        parser.add_argument('-n', '--name', type=str, default=None, )
        parser.add_argument('-d', '--dry', action='store_true', )
        parser.add_argument('-f', '--dry-full', action='store_true', )
        parser.add_argument('-p', '--pattern', type=str, default=None)
        parser.add_argument('-pa', '--pattern-append', type=str, default=None, help=Help.pattern_append)


import re
from pprint import pprint as pp


class DefaultHelp(AppArgument):
    target = 'primary'

    def setup_args(self, parser):
        # parser.add_help = False
        del parser._option_string_actions['--help']
        del parser._option_string_actions['-h']
        parser.add_argument('-h', '--help', action='help')
        # parser.set_defaults(func=self.caller)

    def caller(self, *args, **kwargs):
        print('HELP', args, kwargs)

    # def prep(self, parser):
    #     print('DefaultHelp.prep')
    #     parser.primary_init_kwargs['add_help'] = False
    #     return super().prep(parser)


class ScriptInstall(ConfigMixin):
    """Install a script given through the CLI by file.
    """

    def parse_script_path_place(self, parsed, **kwargs):
        return parsed.filename.absolute().as_posix()

    def parse_raw_path_place(self, parsed, **kwargs):
        return str(parsed.filename)

    def as_path(self, parts):
        return Path(''.join(parts))

    def install(self, parsed, **kwargs):
        path = getattr(self, f'parse_{parsed.type}_path_place')(parsed, **kwargs)

        name = parsed.name or self.as_path(parsed.filename).stem
        parts = self.parse_name(name, parsed.pattern_append, parsed.pattern)
        print('Installing', name)
        data = self.get_conf()
        if parsed.dry:
            r = self.append_graph(data if parsed.dry_full else {}, parts, name)
            r[name] = path
            pp(r)
            return
        self.append_graph(data, parts, name)
        data[name] = path
        self.write_conf_data(data)

    def append_graph(self, data, parts, name, graph=None):

        graph = graph or data.get('graph', None)
        if 'graph' not in data:
            data['graph'] = { 'keys': []}
        leaf = data['graph']

        for i, part in enumerate(parts):
            if part not in leaf:
                print('Add leaf', part)
                leaf[part] = {'depth': i, 'keys': [], 'position': parts[0: i] }
            keys = set(leaf['keys'])
            keys.add(part)
            leaf['keys'] = tuple(keys)
            leaf = leaf[part]

        items = set(leaf.get('items') or [])
        items.add(name)
        leaf['items'] = tuple(items)
        return data

    def parse_name(self, name, append_pattern='', pattern=None):
        safe_ap = append_pattern or ''
        pattern = pattern or f"[_ -{safe_ap}]"
        return re.split(pattern, name)


class ScriptsAdd(AppFunction):
    """The "scripts add [filename]" command for installing a command into
    the persistent store.
    """
    parent_name = 'scripts'
    name = 'add'
    help = Help.add
    register_name ='admin'

    arguments = (ScriptsAddFilenameArg,)

    def hook(self, parsed, *args, **kwargs):
        print('hook', parsed)
        si = ScriptInstall()
        si.install(parsed, **kwargs)
        return 'res from hook'


actions = TrimApp()
# actions.add(VerboseSwitch, VersionSwitch, Scripts, ScriptsAdd, ScriptsAddFilenameArg)
actions.prep()

admin_actions = TrimAdminApp()
admin_actions.add(AppActions, Scripts)
# admin_actions.add(VerboseSwitch, VersionSwitch, Scripts, ScriptsAdd, ScriptsAddFilenameArg)
admin_actions.prep()


def main():
    print('Entry from main')
    actions.run_hook()


def main_admin():
    print('Admin entry from main')
    admin_actions.run_hook()



from pathlib import Path
from django.http import FileResponse

def content_type_response(filepath, ext=None, content_types=None, default=None):
    default = default or 'application/octet-stream'

    content_type_map = content_types or {
        'png': 'image/png',
        'pdf': "application/pdf",
        'unknown': default,
    }

    content_type = content_type_map.get(Path(filepath).suffix[1:], default)

    response = FileResponse(filepath.open('rb'),
                            content_type=content_type)

    response["Content-Disposition"] = "filename={}".format(filepath.name)
    return response



import shlex, subprocess
import sys


def read_one_stream_command(command):
    """Open a cli command, executing the task waiting for a returncode completion
      Inputs are processed through the pip.

        subprocess.STD_INPUT_HANDLE
            The standard input device. Initially, this is the console input
            buffer, CONIN$.
        subprocess.STD_OUTPUT_HANDLE
            The standard output device. Initially, this is the active console
            screen buffer, CONOUT$.
        subprocess.STD_ERROR_HANDLE
            The standard error device. Initially, this is the active console
            screen buffer, CONOUT$.
        subprocess.SW_HIDE
            Hides the window. Another window will be activated.
        subprocess.STARTF_USESTDHANDLES
            Specifies that the STARTUPINFO.hStdInput, STARTUPINFO.hStdOutput,
            and STARTUPINFO.hStdError attributes contain additional information.
        subprocess.STARTF_USESHOWWINDOW
            Specifies that the STARTUPINFO.wShowWindow attribute contains
            additional information.
        subprocess.CREATE_NEW_CONSOLE
            The new process has a new console, instead of inheriting its
            parent’s console (the default).
        subprocess.CREATE_NEW_PROCESS_GROUP
            A Popen creationflags parameter to specify that a new process
            group will be created. This flag is necessary for using os.kill()
            on the subprocess.
            This flag is ignored if CREATE_NEW_CONSOLE is specified.
        subprocess.ABOVE_NORMAL_PRIORITY_CLASS
            A Popen creationflags parameter to specify that a new process will
            have an above average priority.
        subprocess.BELOW_NORMAL_PRIORITY_CLASS
            A Popen creationflags parameter to specify that a new process will
            have a below average priority.
        subprocess.HIGH_PRIORITY_CLASS
            A Popen creationflags parameter to specify that a new process will
            have a high priority.
        subprocess.IDLE_PRIORITY_CLASS
            A Popen creationflags parameter to specify that a new process will
            have an idle (lowest) priority.
        subprocess.NORMAL_PRIORITY_CLASS
            A Popen creationflags parameter to specify that a new process will
            have an normal priority. (default)
        subprocess.REALTIME_PRIORITY_CLASS
            A Popen creationflags parameter to specify that a new process will
            have realtime priority. You should almost never use
            REALTIME_PRIORITY_CLASS, because this interrupts system threads
            that manage mouse input, keyboard input, and background disk
            flushing. This class can be appropriate for applications that
            “talk” directly to hardware or that perform brief tasks that
            should have limited interruptions.
        subprocess.CREATE_NO_WINDOW
            A Popen creationflags parameter to specify that a new process will
            not create a window.
        subprocess.DETACHED_PROCESS
            A Popen creationflags parameter to specify that a new process will
            not inherit its parent’s console. This value cannot be
            used with CREATE_NEW_CONSOLE.
        subprocess.CREATE_DEFAULT_ERROR_MODE
            A Popen creationflags parameter to specify that a new process does
            not inherit the error mode of the calling process. Instead, the
            new process gets the default error mode. This feature is
            particularly useful for multithreaded shell applications that
            run with hard errors disabled.
        subprocess.CREATE_BREAKAWAY_FROM_JOB
            A Popen creationflags parameter to specify that a new process is
            not associated with the job.
    """
    # si = subprocess.STARTUPINFO()
    flags = 0 # subprocess.DETACHED_PROCESS

    process = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        creationflags=flags,
        ## not required when processing by read(1)
        # bufsize=1,
        ## Adding universal newlines will evoke _str_ from the `out`
        # universal_newlines=True,
    )

    byte_count = 0
    count = 0
    while True:
        out = process.stdout.read(1)
        if out == '' and process.poll() != None:
            break
        byte_count += 1

        if byte_count > 75_000:
            break

        if len(out) > 0:
            sys.stdout.write(clean(out, ''))
            sys.stdout.flush()
        else:
            count += 1
            if count % 20 == 0:
                print('blanked.')
                break
    return proc_wait(process)


def proc_wait(proc, timeout=10):
    try:
        outs, errs = proc.communicate(timeout=timeout)
    except subprocess.TimeoutExpired:
        proc.kill()
        outs, errs = proc.communicate()
    return outs, errs


def clean(text, default=None):
    t=text
    if isinstance(text, bytes):
        t = text.decode('utf')
    return default if len(t) == 0 else t



"""
The "live" models tool provides a fast access method to all the existing
installed models using a single concat string, resolving to the model.

This can replace a standard model import, or the django get_model import:


with get_model:

    from django.apps import apps

    ShippingAccount = apps.get_model('baskets.ShippingAccount')
    cart_owner = self.get_object().cart.owner
    user_addresses = ShippingAccount.objects.filter(owner=cart_owner, deleted=False)

after:

    from trim import live

    owner = self.get_object().cart.owner
    user_addresses = live.baskets.ShippingAccount.objects.filter(
            owner=owner, deleted=False
        )
"""
from django.apps import apps


class MagicModelApp(object):
    """The top level 'live' model import, expecting a string caller:

        live = MagicModelApp()
        live.appname

    """
    def __getattr__(self, k):
        return MagicModelModel(k)


class MagicModelModel(object):
    """A return value from the live appname collection, returning
    a model upon a key call:

        live.appname.ModelName
        <appname.ModelName>
    """
    def __init__(self, appname):
        self.appname = appname

    def __getattr__(self, k):
        return apps.get_model(self.appname, k)


# class MagicModelProps(object):

#     def __init__(self, appname, modelname):
#         self.appname = appname
#         self.modelname = modelname

#     def __getattr__(self, k):

live = MagicModelApp()
